import asyncio
from time import time
import aiohttp
import async_timeout

from .channel import BaseChannel
from .settings import conf

class PM(BaseChannel):
    """Manages a connection with Chatango PM."""

    event_name_ping = 'onPMPing'

    user_class = None

    contacts = None
    blocklist = None

    _status = None

    def __init__(self, *args, **kwargs):
        super(PM, self).__init__(*args, **kwargs)
        self.user_class = self.mgr.user_class

        self.blocklist = set()
        self.contacts = set()
        self._status = {}


    def get_server(self):
        return self.mgr.get_pm_host()


    def _disconnect(self):
        self._call_event('onPMDisconnect')


    @asyncio.coroutine
    def authenticate(self):
        try:
            username, password = self.mgr.get_user_credentials()
        except ValueError:
            return

        auth_token = yield from self._get_auth_token(username, password)
        if auth_token:
            self._send_command('tlogin', auth_token, '2')
        else:
            self._call_event('onLoginFail')
            self.disconnect()


    @asyncio.coroutine
    def _get_auth_token(self, name, password):
        """
        Request an auid using name and password.

        @type name: str
        @param name: name
        @type password: str
        @param password: password

        @rtype: str
        @return: auid
        """
        payload = {
            'user_id': name,
            'password': password,
            'storecookie': 'on',
            'checkerrors': 'yes',
        }
        token = None
        session = aiohttp.ClientSession(loop=self._loop)
        try:
            with async_timeout.timeout(conf['connection']['timeout']):
                resp = yield from session.post('https://chatango.com/login',
                        data=payload)

                token = resp.cookies['auth.chatango.com'].value
                yield from resp.release()
        except asyncio.TimeoutError:
            pass
        except: # pylint:disable=bare-except
            self._log.exception('PM::_get_auth_token')
        finally:
            session.close()

        return token


    @asyncio.coroutine
    def message(self, user, msg):
        """send a pm to a user"""
        if msg is not None:
            self._send_command('msg', user.name, msg)


    def addContact(self, user):
        """add contact"""
        if user not in self.contacts:
            self._send_command('wladd', user.name)
            self.contacts.add(user)
            self._call_event('onPMContactAdd', user)


    def removeContact(self, user):
        """remove contact"""
        if user in self.contacts:
            self._send_command('wldelete', user.name)
            self.contacts.remove(user)
            self._call_event('onPMContactRemove', user)


    def block(self, user):
        """block a person"""
        if user not in self.blocklist:
            self._send_command('block', user.name, user.name, 'S')
            self.blocklist.add(user)
            self._call_event('onPMBlock', user)


    def unblock(self, user):
        """unblock a person"""
        if user in self.blocklist:
            self._send_command('unblock', user.name)


    def track(self, user):
        """get and store status of person for future use"""
        self._send_command('track', user.name)


    def checkOnline(self, user):
        """return True if online, False if offline, None if unknown"""
        if user in self._status:
            return self._status[user][1]
        else:
            return None


    def getIdle(self, user):
        """
        return last active time, time.time() if isn't idle, 0 if offline,
        None if unknown
        """
        if not user in self._status:
            return None
        elif not self._status[user][1]:
            return 0
        elif not self._status[user][2]:
            return time()
        else:
            return self._status[user][2]


    ## Received commands


    @asyncio.coroutine
    def _rcmd_OK(self, args):
        self._send_command('wl')
        self._send_command('getblock')
        self._call_event('onPMConnect')


    @asyncio.coroutine
    def _rcmd_wl(self, args):
        self.contacts = set()
        for i in range(len(args) // 4):
            name, last_on, is_on, idle = args[i * 4: i * 4 + 4]
            user = self.user_class.create(name)
            if last_on is "None":
                #in case chatango gives a "None" as data argument
                pass
            elif not is_on == "on":
                self._status[user] = [int(last_on), False, 0]
            elif idle == '0':
                self._status[user] = [int(last_on), True, 0]
            else:
                self._status[user] = [int(last_on), True,
                        time() - int(idle) * 60]

            self.contacts.add(user)

        self._call_event('onPMContactlistReceive')


    @asyncio.coroutine
    def _rcmd_block_list(self, args):
        self.blocklist = set()
        for name in args:
            if name == "":
                continue
            self.blocklist.add(self.user_class.create(name))


    @asyncio.coroutine
    def _rcmd_idleupdate(self, args):
        user = self.user_class.create(args[0])
        try:
            # last_on, is_on, idle
            last_on, is_on, _ = self._status[user]
        except: # pylint: disable=bare-except
            pass

        if args[1] == '1':
            self._status[user] = [last_on, is_on, 0]
        else:
            try:
                self._status[user] = [last_on, is_on, time()]
            except: # pylint: disable=bare-except
                pass


    @asyncio.coroutine
    def _rcmd_track(self, args):
        user = self.user_class.create(args[0])
        if user in self._status:
            last_on = self._status[user][0]
        else:
            last_on = 0

        if args[1] == '0':
            idle = 0
        else:
            idle = time() - int(args[1]) * 60

        is_on = args[2] == "online"

        self._status[user] = [last_on, is_on, idle]


    @asyncio.coroutine
    def _rcmd_DENIED(self, args):
        self._call_event('onLoginFail')
        self.disconnect()


    @asyncio.coroutine
    def _rcmd_msg(self, args):
        user = self.user_class.create(args[0])
        body = self.strip_tags(':'.join(args[5:]))
        self._call_event('onPMMessage', user, body)


    @asyncio.coroutine
    def _rcmd_msgoff(self, args):
        user = self.user_class.create(args[0])
        body = self.strip_tags(':'.join(args[5:]))
        self._call_event('onPMOfflineMessage', user, body)


    @asyncio.coroutine
    def _rcmd_wlonline(self, args):
        user = self.user_class.create(args[0])
        last_on = float(args[1])
        self._status[user] = [last_on, True, last_on]
        self._call_event('onPMContactOnline', user)


    @asyncio.coroutine
    def _rcmd_wloffline(self, args):
        user = self.user_class.create(args[0])
        last_on = float(args[1])
        self._status[user] = [last_on, False, 0]
        self._call_event('onPMContactOffline', user)


    @asyncio.coroutine
    def _rcmd_kickingoff(self, args):
        self.disconnect()


    @asyncio.coroutine
    def _rcmd_toofast(self, args):
        self.disconnect()


    @asyncio.coroutine
    def _rcmd_unblocked(self, args):
        """call when successfully unblocked"""
        user = self.user_class.create(args[0])
        if user in self.blocklist:
            self.blocklist.remove(user)
            self._call_event('onPMUnblock', user)



from collections import Mapping
from copy import deepcopy
from json import load as json_load
import logging
import os

DEFAULT_SETTINGS = {
    'authentication': {
        'username': None,
        'password': None,
    },
    'datastore': {
        'connect_args': {},
        'prefix': 'chatango_bot_', # prefix of stored data's dictionary key
    },
    'connection': {
        'buffer_size': 65536,
        'max_retries': 99,
        'ping_interval': 20, # seconds
        'retry_delay': 10, # seconds
        'timeout': 30, # seconds
    },
    'servers': {
        'anonymous_pm_host': 'b1.chatango.com',
        'anonymous_pm_port': 5222,
        'pm_host': 'c1.chatango.com',
        'pm_port': 5222,
        'chatroom_port': 443,
    },
    'logging': {
        'level': 'info',
    },
    'message_formatter': {
        'max_length': 1800,
        'overflow': 'chunked', # too large, chunked or crop?
    },
    'user_list': {
        'active_filter': 'recent',
        'unique': True,

        'filters': {
            'recent': {
                'size': 50,
            },
        },
    },
    'history': {
        'size': 150,
    },
    'chat': {
        'name_color': '000099',
        'text_color': '000099',
        'text_typeface': '1',
        'text_size': 11,
    },
    'services': [],
}

def _deep_update(obj, data):
    for key in data:
        item = data[key]

        if isinstance(item, Mapping):
            if key not in obj:
                obj[key] = {}
            _deep_update(obj[key], item)
        else:
            obj[key] = item


conf = deepcopy(DEFAULT_SETTINGS)

with open(os.environ['SETTINGS_FILE'], 'r') as f:
    _deep_update(conf, json_load(f))


logging.basicConfig(level=getattr(logging, conf['logging']['level'].upper()))



import asyncio
from logging import getLogger

from .channel import BaseChannel

class AnonPM(BaseChannel):
    """Manages connection with Chatango anon PM."""

    event_name_ping = 'onPMPing'

    user_class = None

    def __init__(self, name, *args, **kwargs):
        self.name = name
        super(AnonPM, self).__init__(*args, **kwargs)
        self.user_class = self.mgr.user_class


    def get_server(self):
        return self.mgr.get_anonpm_host()


    def _disconnect(self):
        self._call_event('onAnonPMDisconnect',
                self.user_class.create(self.name))


    @asyncio.coroutine
    def authenticate(self):
        self._send_command('mhs', 'mini', 'unknown', self.name)


    @asyncio.coroutine
    def message(self, user, msg):
        """send a pm to a user"""
        if msg is not None:
            self._send_command('msg', user.name, msg)


    ## Received Commands


    @asyncio.coroutine
    def _rcmd_mhs(self, args):
        """
        note to future maintainers

        args[1] is ether "online" or "offline"
        """
        self.disconnect()


    @asyncio.coroutine
    def _rcmd_msg(self, args):
        user = self.user_class.create(args[0])
        body = self.strip_tags(':'.join(args[5:]))
        self._call_event('onPMMessage', user, body)


class AnonPMManager(object):
    """Comparable wrapper for anon Chatango PM"""

    channel_class = AnonPM

    mgr = None

    _channels = None

    _log = None
    _loop = None

    def __init__(self, loop, mgr):
        self.mgr = mgr
        self._log = getLogger('AnonPMManager')
        self._loop = loop

        self._channels = {}


    @asyncio.coroutine
    def connect(self, username):
        channel = self.channel_class(loop=self._loop, mgr=self.mgr,
                name=username)

        self._channels[username.lower()] = channel

        yield from channel.connect()


    def disconnect(self):
        tasks = []

        for channel in self._channels.values():
            future = channel.disconnect()
            if future is not None:
                tasks.append(future)

        future = asyncio.gather(*tasks, loop=self._loop, return_exceptions=True)
        return future


    @asyncio.coroutine
    def message(self, user, msg):
        """send a pm to a user"""
        username = user.name.lower()

        if not username in self._channels:
            yield from self.connect(user.name)
            yield from asyncio.sleep(5)

        yield from self._channels[username].message(user, msg)



class Message(object):
    """Class that represents a message."""

    msgid = None
    time = None
    user = None
    body = None
    room = None
    raw = ''
    ip = None
    unid = ''
    nameColor = '000'
    fontSize = 12
    fontFace = '0'
    fontColor = '000'

    @property
    def id(self):
        return self.msgid


    def attach(self, room, msgid):
        """
        Attach the Message to a message id.

        @type msgid: str
        @param msgid: message id
        """
        if self.msgid is None:
            self.room = room
            self.msgid = msgid
            self.room._msgs[msgid] = self


    def detach(self):
        """Detach the Message."""

        if self.msgid is not None and self.msgid in self.room._msgs:
            del self.room._msgs[self.msgid]
            self.msgid = None


    def delete(self):
        self.room.deleteMessage(self)


    def __init__(self, **kw):
        for attr, val in kw.items():
            if val is not None:
                setattr(self, attr, val)



import asyncio
from logging import getLogger

from .settings import conf
from .room import Room
from .user import User
from .pm import PM
from .anonpm import AnonPMManager
from .message import Message

class Manager(object):
    """Class that manages multiple connections."""

    room_class = Room
    user_class = User
    pm_class = PM
    anonpm_class = AnonPMManager
    message_class = Message

    user = None
    rooms = None
    pm = None

    _loop = None
    _log = None
    _future = None

    @property
    def roomnames(self):
        return [room.name for room in self.rooms.values()]


    def __init__(self, loop, pm=True):
        self._loop = loop
        self._log = getLogger(type(self).__name__)

        self.rooms = {}
        self.user = self.user_class.create(conf['authentication']['username'])

        if pm:
            _, password = self.get_user_credentials()
            if password:
                self.pm = self.pm_class(loop=loop, mgr=self)
            else:
                self.pm = self.anonpm_class(loop=loop, mgr=self) # pylint: disable=redefined-variable-type

        asyncio.ensure_future(self.onInit())


    def get_room_host(self, room_name):
        """
        Get the server host for a certain room.

        @type group: str
        @param group: room name

        @rtype: str
        @return: the server's hostname
        """
        specials = {
            'mitvcanal': 56,
            'animeultimacom': 34,
            'cricket365live': 21,
            'pokemonepisodeorg': 22,
            'animelinkz': 20,
            'sport24lt': 56,
            'narutowire': 10,
            'watchanimeonn': 22,
            'cricvid-hitcric-': 51,
            'narutochatt': 70,
            'leeplarp': 27,
            'stream2watch3': 56,
            'ttvsports': 56,
            'ver-anime': 8,
            'vipstand': 21,
            'eafangames': 56,
            'soccerjumbo': 21,
            'myfoxdfw': 67,
            'kiiiikiii': 21,
            'de-livechat': 5,
            'rgsmotrisport': 51,
            'dbzepisodeorg': 10,
            'watch-dragonball': 8,
            'peliculas-flv': 69,
            'tvanimefreak': 54,
            'tvtvanimefreak': 54,
        }
        tsweights = (
            ('5', 75),
            ('6', 75),
            ('7', 75),
            ('8', 75),
            ('16', 75),
            ('17', 75),
            ('18', 75),
            ('9', 95),
            ('11', 95),
            ('12', 95),
            ('13', 95),
            ('14', 95),
            ('15', 95),
            ('19', 110),
            ('23', 110),
            ('24', 110),
            ('25', 110),
            ('26', 110),
            ('28', 104),
            ('29', 104),
            ('30', 104),
            ('31', 104),
            ('32', 104),
            ('33', 104),
            ('35', 101),
            ('36', 101),
            ('37', 101),
            ('38', 101),
            ('39', 101),
            ('40', 101),
            ('41', 101),
            ('42', 101),
            ('43', 101),
            ('44', 101),
            ('45', 101),
            ('46', 101),
            ('47', 101),
            ('48', 101),
            ('49', 101),
            ('50', 101),
            ('52', 110),
            ('53', 110),
            ('55', 110),
            ('57', 110),
            ('58', 110),
            ('59', 110),
            ('60', 110),
            ('61', 110),
            ('62', 110),
            ('63', 110),
            ('64', 110),
            ('65', 110),
            ('66', 110),
            ('68', 95),
            ('71', 116),
            ('72', 116),
            ('73', 116),
            ('74', 116),
            ('75', 116),
            ('76', 116),
            ('77', 116),
            ('78', 116),
            ('79', 116),
            ('80', 116),
            ('81', 116),
            ('82', 116),
            ('83', 116),
            ('84', 116),
        )
        try:
            sn = specials[room_name]
        except KeyError:
            group = room_name.replace('_', 'q').replace('-', 'q')
            fnv = float(int(group[0:min(5, len(group))], 36))
            lnv = group[6: (6 + min(3, len(group) - 5))]
            if(lnv):
                lnv = float(int(lnv, 36))
                lnv = max(lnv, 1000)
            else:
                lnv = float(1000)
            num = (fnv % lnv) / lnv
            maxnum = sum(map(lambda x: x[1], tsweights))
            cumfreq = 0
            sn = 0
            for wgt in tsweights:
                cumfreq += float(wgt[1]) / maxnum
                if(num <= cumfreq):
                    sn = int(wgt[0])
                    break

        host = 's%s.chatango.com' % sn
        port = conf['servers']['chatroom_port']
        return (host, port)


    def get_anonpm_host(self):
        host = conf['servers']['anonymous_pm_host']
        port = conf['servers']['anonymous_pm_port']
        return (host, port)


    def get_pm_host(self):
        host = conf['servers']['pm_host']
        port = conf['servers']['pm_port']
        return (host, port)


    def get_user_credentials(self):
        username = conf['authentication']['username']
        password = conf['authentication']['password']
        return (username, password)


    @asyncio.coroutine
    def joinRoom(self, room_name):
        """
        Join a room or return None if already joined.

        @type room: str
        @param room: room to join

        @rtype: Room or None
        @return: True or nothing
        """
        room_name = room_name.lower()
        room = self.rooms.get(room_name)
        if room is None:
            room = self.room_class(name=room_name, loop=self._loop, mgr=self)
            self.rooms[room_name] = room

        if not room.connected:
            yield from room.connect()

        return room


    @asyncio.coroutine
    def leaveRoom(self, room_name):
        """
        Leave a room.

        @type room: str
        @param room: room to leave
        """
        try:
            room = self.rooms.pop(room_name.lower())
            yield from room.disconnect()
        except KeyError:
            pass


    def getRoom(self, room_name):
        """
        Get room with a name, or None if not connected to this room.

        @type room: str
        @param room: room

        @rtype: Room
        @return: the room
        """
        return self.rooms.get(room_name.lower())


    def connect(self, *room_names):
        if self._future is None:
            try:
                self._future = self._loop.create_future()
            except AttributeError:
                self._future = asyncio.Future()

        if isinstance(self.pm, self.pm_class):
            asyncio.ensure_future(self.pm.connect()) # pylint:disable=no-value-for-parameter

        for room_name in room_names:
            if hasattr(room_name, 'decode'):
                room_name = room_name.decode('utf-8')
            asyncio.ensure_future(self.joinRoom(room_name))

        return self._future


    def disconnect(self):
        tasks = []
        if self.pm:
            future = self.pm.disconnect()
            if future is not None:
                tasks.append(future)

        for room in self.rooms.values():
            future = room.disconnect()
            if future is not None:
                tasks.append(future)

        future = asyncio.gather(*tasks, loop=self._loop, return_exceptions=True)
        future.add_done_callback(self._future.set_result)

        result = self._future
        self._future = None
        return result


    ## Commands


    def enableBg(self):
        """Enable background if available."""
        self.user._mbg = True
        for room in self.rooms.values():
            room.setBgMode(1)


    def disableBg(self):
        """Disable background."""
        self.user._mbg = False
        for room in self.rooms.values():
            room.setBgMode(0)


    def enableRecording(self):
        """Enable recording if available."""
        self.user._mrec = True
        for room in self.rooms.values():
            room.setRecordingMode(1)


    def disableRecording(self):
        """Disable recording."""
        self.user._mrec = False
        for room in self.rooms.values():
            room.setRecordingMode(0)


    def setNameColor(self, rgb_hex):
        """
        Set name color.

        @type rgb_hex: str
        @param rgb_hex: a 3-char RGB hex code for the color
        """
        self.user.nameColor = rgb_hex


    def setFontColor(self, rgb_hex):
        """
        Set font color.

        @type rgb_hex: str
        @param rgb_hex: a 3-char RGB hex code for the color
        """
        self.user.fontColor = rgb_hex


    def setFontFace(self, face):
        """
        Set font face/family.

        @type face: str
        @param face: the font face
        """
        self.user.fontFace = face


    def setFontSize(self, size):
        """
        Set font size.

        @type size: int
        @param size: the font size (limited: 9 to 22)
        """
        if size < 9:
            size = 9
        if size > 22:
            size = 22
        self.user.fontSize = size


    ## Virtual methods


    @asyncio.coroutine
    def onInit(self):
        """Called on init."""
        pass


    @asyncio.coroutine
    def onConnect(self, room):
        """
        Called when connected to the room.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onReconnect(self, room):
        """
        Called when reconnected to the room.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onConnectFail(self, room):
        """
        Called when the connection failed.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onDisconnect(self, room):
        """
        Called when the client gets disconnected.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onLoginFail(self, room):
        """
        Called on login failure, disconnects after.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onFloodBan(self, room):
        """
        Called when either flood banned or flagged.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onFloodBanRepeat(self, room):
        """
        Called when trying to send something when floodbanned.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onFloodWarning(self, room):
        """
        Called when an overflow warning gets received.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onMessageDelete(self, room, user, message):
        """
        Called when a message gets deleted.

        @type room: Room
        @param room: room where the event occured
        @type user: User
        @param user: owner of deleted message
        @type message: Message
        @param message: message that got deleted
        """
        pass


    @asyncio.coroutine
    def onModChange(self, room):
        """
        Called when the moderator list changes.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onModAdd(self, room, user):
        """
        Called when a moderator gets added.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onModRemove(self, room, user):
        """
        Called when a moderator gets removed.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onMessage(self, room, user, message):
        """
        Called when a message gets received.

        @type room: Room
        @param room: room where the event occured
        @type user: User
        @param user: owner of message
        @type message: Message
        @param message: received message
        """
        pass


    @asyncio.coroutine
    def onHistoryMessage(self, room, user, message):
        """
        Called when a message gets received from history.

        @type room: Room
        @param room: room where the event occured
        @type user: User
        @param user: owner of message
        @type message: Message
        @param message: the message that got added
        """
        pass


    @asyncio.coroutine
    def onJoin(self, room, user):
        """
        Called when a user joins. Anonymous users get ignored here.

        @type room: Room
        @param room: room where the event occured
        @type user: User
        @param user: the user that has joined
        """
        pass


    @asyncio.coroutine
    def onLeave(self, room, user):
        """
        Called when a user leaves. Anonymous users get ignored here.

        @type room: Room
        @param room: room where the event occured
        @type user: User
        @param user: the user that has left
        """
        pass


    @asyncio.coroutine
    def onRaw(self, room, raw):
        """
        Called before any command parsing occurs.

        @type room: Room
        @param room: room where the event occured
        @type raw: str
        @param raw: raw command data
        """
        pass


    @asyncio.coroutine
    def onPing(self, room):
        """
        Called when a ping gets sent.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onUserCountChange(self, room):
        """
        Called when the user count changes.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onBan(self, room, user, target):
        """
        Called when a user gets banned.

        @type room: Room
        @param room: room where the event occured
        @type user: User
        @param user: user that banned someone
        @type target: User
        @param target: user that got banned
        """
        pass


    @asyncio.coroutine
    def onUnban(self, room, user, target):
        """
        Called when a user gets unbanned.

        @type room: Room
        @param room: room where the event occured
        @type user: User
        @param user: user that unbanned someone
        @type target: User
        @param target: user that got unbanned
        """
        pass


    @asyncio.coroutine
    def onBanlistUpdate(self, room):
        """
        Called when a banlist gets updated.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onUnBanlistUpdate(self, room):
        """
        Called when a unbanlist gets updated.

        @type room: Room
        @param room: room where the event occured
        """
        pass


    @asyncio.coroutine
    def onPMConnect(self, pm):
        """
        Called when connected to the pm

        @type pm: PM
        @param pm: the pm
        """
        pass


    @asyncio.coroutine
    def onAnonPMDisconnect(self, pm, user):
        """
        Called when disconnected from the pm

        @type pm: PM
        @param pm: the pm
        """
        pass


    @asyncio.coroutine
    def onPMDisconnect(self, pm):
        """
        Called when disconnected from the pm

        @type pm: PM
        @param pm: the pm
        """
        pass


    @asyncio.coroutine
    def onPMPing(self, pm):
        """
        Called when sending a ping to the pm

        @type pm: PM
        @param pm: the pm
        """
        pass


    @asyncio.coroutine
    def onPMMessage(self, pm, user, body):
        """
        Called when a message is received

        @type pm: PM
        @param pm: the pm
        @type user: User
        @param user: owner of message
        @type message: Message
        @param message: received message
        """
        pass


    @asyncio.coroutine
    def onPMOfflineMessage(self, pm, user, body):
        """
        Called when connected if a message is received while offline

        @type pm: PM
        @param pm: the pm
        @type user: User
        @param user: owner of message
        @type message: Message
        @param message: received message
        """
        pass


    @asyncio.coroutine
    def onPMContactlistReceive(self, pm):
        """
        Called when the contact list is received

        @type pm: PM
        @param pm: the pm
        """
        pass


    @asyncio.coroutine
    def onPMBlocklistReceive(self, pm):
        """
        Called when the block list is received

        @type pm: PM
        @param pm: the pm
        """
        pass


    @asyncio.coroutine
    def onPMContactAdd(self, pm, user):
        """
        Called when the contact added message is received

        @type pm: PM
        @param pm: the pm
        @type user: User
        @param user: the user that gotten added
        """
        pass


    @asyncio.coroutine
    def onPMContactRemove(self, pm, user):
        """
        Called when the contact remove message is received

        @type pm: PM
        @param pm: the pm
        @type user: User
        @param user: the user that gotten remove
        """
        pass


    @asyncio.coroutine
    def onPMBlock(self, pm, user):
        """
        Called when successfully block a user

        @type pm: PM
        @param pm: the pm
        @type user: User
        @param user: the user that gotten block
        """
        pass


    @asyncio.coroutine
    def onPMUnblock(self, pm, user):
        """
        Called when successfully unblock a user

        @type pm: PM
        @param pm: the pm
        @type user: User
        @param user: the user that gotten unblock
        """
        pass


    @asyncio.coroutine
    def onPMContactOnline(self, pm, user):
        """
        Called when a user from the contact come online

        @type pm: PM
        @param pm: the pm
        @type user: User
        @param user: the user that came online
        """
        pass


    @asyncio.coroutine
    def onPMContactOffline(self, pm, user):
        """
        Called when a user from the contact go offline

        @type pm: PM
        @param pm: the pm
        @type user: User
        @param user: the user that went offline
        """
        pass


    @asyncio.coroutine
    def onEventCalled(self, room, evt, *args, **kw):
        """
        Called on every room-based event.

        @type room: Room
        @param room: room where the event occured
        @type evt: str
        @param evt: the event
        """
        pass


    @asyncio.coroutine
    def onUserList(self, room, user_list):
        """
        Called when bot connected and given room members.

        @type room: Room
        @param room: room where the event occured
        @type user_list: List of User
        @param user_list: users inside the room
        """
        pass



import asyncio
import random

from time import time

from .channel import BaseChannel
from .settings import conf

ROOM_OWNER = 2
ROOM_MODERATOR = 1

class Struct(object):
    def __init__(self, **entries):
        self.__dict__.update(entries)


class Room(BaseChannel):
    """Manages a connection with a Chatango room."""

    user_class = None
    message_class = None

    owner = None
    usercount = 0
    silent = False
    mods = None

    _botname = None
    _history = None
    _userlist = None
    _banlist = None
    _unbanlist = None
    _mqueue = None
    _connectAmmount = 0
    _premium = False
    _msgs = None
    _i_log = None

    _uid = None
    _aid = None


    def __init__(self, name, *args, **kwargs):
        self.name = name
        super(Room, self).__init__(*args, **kwargs)
        self.user_class = self.mgr.user_class
        self.message_class = self.mgr.message_class

        self.mods = set()

        self._mqueue = {}
        self._history = []
        self._userlist = []
        self._msgs = {}
        self._banlist = {}
        self._unbanlist = {}
        self._i_log = []


    def get_server(self):
        return self.mgr.get_room_host(self.name)


    def _disconnect(self):
        self._call_event('onDisconnect')


    @asyncio.coroutine
    def authenticate(self):
        """Chatango authentication."""
        try:
            username, password = self.mgr.get_user_credentials()
            unique_id = str(random.randrange(10 ** 15, 10 ** 16))
            self._send_command('bauth', self.name, unique_id, username,
                    password)

        except ValueError:
            # Login as anon
            self._send_command('bauth', self.name)


    @property
    def botname(self):
        try:
            username, password = self.mgr.get_user_credentials()

            if password is None:
                return '#' + username
            else:
                return username
        except ValueError:
            return self._botname


    @property
    def userlist(self):
        return self._get_userlist()


    @property
    def usernames(self):
        return [user.name for user in self._userlist]


    @property
    def user(self):
        return self.mgr.user


    @property
    def username(self):
        return self.user.name


    @property
    def ownername(self):
        return self.owner.name


    @property
    def modnames(self):
        return [user.name for user in self.mods]


    @property
    def banlist(self):
        return self._banlist.keys()


    @property
    def unbanlist(self):
        return [(record["target"], record["src"]) for record in\
                self._unbanlist.values()]


    def _get_userlist(self, mode=None, unique=None, memory=None):
        ul = None
        if mode is None:
            mode = conf['user_list']['active_filter']

        if unique is None:
            unique = conf['user_list']['unique']

        if memory is None:
            memory = conf['user_list']['filters']['recent']['size']

        if mode == 'recent':
            ul = [hist.user for hist in self._history[-memory:]]
        else:
            ul = self._userlist

        if unique:
            return list(set(ul))
        else:
            return ul


    def disconnect(self):
        future = super(Room, self).disconnect()

        for user in self._userlist:
            user.clearSessionIds(self)

        self._userlist = []
        return future



    def login(self, username, password=None):
        """login as a user or set a name in room"""
        if password is not None:
            self._send_command('blogin', username, password)
        else:
            self._send_command('blogin', username)


    def logout(self):
        """logout of user in a room"""
        self._send_command('blogout')


    def rawMessage(self, msg):
        """
        Send a message without n and f tags.

        @type msg: str
        @param msg: message
        """
        if not self.silent:
            self._send_command('bmsg:tl2r', msg)


    @asyncio.coroutine
    def message(self, msg, html=False):
        """
        Send a message. (Use "\n" for new line)

        @type msg: str
        @param msg: message
        """
        if msg is None:
            return
        msg = msg.rstrip()
        if not html:
            msg = self.html_escape(msg)

        max_length = conf['message_formatter']['max_length']
        overflow = conf['message_formatter']['overflow']

        if len(msg) > max_length:
            if overflow == 'crop':
                yield from self.message(msg[:max_length], html=html)
            elif overflow == 'chunked':
                while len(msg) > 0:
                    sect = msg[:max_length]
                    msg = msg[max_length:]
                    yield from self.message(sect, html=html)
            else:
                raise RuntimeError('Unknown message_formatter: ' + overflow)
            return

        msg = '<n' + self.user.nameColor + '/>' + msg
        if self.botname is not None and not self.botname.startswith('!anon'):
            font_properties = '<f x%0.2i%s="%s">' % (self.user.fontSize,
                    self.user.fontColor, self.user.fontFace)

            if '\n' in msg:
                msg = msg.replace('\n', '</f></p><p' + font_properties)
            msg = font_properties + msg

        msg = msg.replace('~', '&#126;')
        self.rawMessage(msg)


    def setBgMode(self, mode):
        """turn on/off bg"""
        self._send_command('msgbg', str(mode))


    def setRecordingMode(self, mode):
        """turn on/off rcecording"""
        self._send_command('msgmedia', str(mode))


    def addMod(self, user):
        """
        Add a moderator.

        @type user: User
        @param user: User to mod.
        """
        if self.getLevel(self.user) == ROOM_OWNER:
            self._send_command('addmod', user.name)


    def removeMod(self, user):
        """
        Remove a moderator.

        @type user: User
        @param user: User to demod.
        """
        if self.getLevel(self.user) == ROOM_OWNER:
            self._send_command('removemod', user.name)


    def flag(self, message):
        """
        Flag a message.

        @type message: Message
        @param message: message to flag
        """
        self._send_command('g_flag', message.msgid)


    def flagUser(self, user):
        """
        Flag a user.

        @type user: User
        @param user: user to flag

        @rtype: bool
        @return: whether a message to flag was found
        """
        msg = self.getLastMessage(user)
        if msg:
            self.flag(msg)
            return True
        return False


    def deleteMessage(self, message):
        """
        Delete a message. (Moderator only)

        @type message: Message
        @param message: message to delete
        """
        if self.getLevel(self.user) >= ROOM_MODERATOR:
            self._send_command('delmsg', message.msgid)


    def deleteUser(self, user):
        """
        Delete a message. (Moderator only)

        @type message: User
        @param message: delete user's last message
        """
        if self.getLevel(self.user) >= ROOM_MODERATOR:
            msg = self.getLastMessage(user)
            if msg:
                self._send_command('delmsg', msg.msgid)
            return True
        return False


    def delete(self, user):
        """
        compatibility wrapper for deleteUser
        """
        self._log.warning('[obsolete] the delete function is obsolete, ' +\
                'please use deleteUser')

        return self.deleteUser(user)


    def rawClearUser(self, unid, ip, user):
        self._send_command('delallmsg', unid, ip, user)


    def clearUser(self, user):
        """
        Clear all of a user's messages. (Moderator only)

        @type user: User
        @param user: user to delete messages of

        @rtype: bool
        @return: whether a message to delete was found
        """
        if self.getLevel(self.user) >= ROOM_MODERATOR:
            msg = self.getLastMessage(user)
            if msg:
                if msg.user.name[0] in ('!', '#'):
                    self.rawClearUser(msg.unid, msg.ip, '')
                else:
                    self.rawClearUser(msg.unid, msg.ip, msg.user.name)

                return True
        return False


    def clearall(self):
        """Clear all messages. (Owner only)"""
        if self.getLevel(self.user) == ROOM_OWNER:
            self._send_command('clearall')


    def rawBan(self, name, ip, unid):
        """
        Execute the block command using specified arguments.
        (For advanced usage)

        @type name: str
        @param name: name
        @type ip: str
        @param ip: ip address
        @type unid: str
        @param unid: unid
        """
        self._send_command('block', unid, ip, name)


    def ban(self, msg):
        """
        Ban a message's sender. (Moderator only)

        @type message: Message
        @param message: message to ban sender of
        """
        if self.getLevel(self.user) >= ROOM_MODERATOR:
            self.rawBan(msg.user.name, msg.ip, msg.unid)


    def banUser(self, user):
        """
        Ban a user. (Moderator only)

        @type user: User
        @param user: user to ban

        @rtype: bool
        @return: whether a message to ban the user was found
        """
        msg = self.getLastMessage(user)
        if msg:
            self.ban(msg)
            return True
        return False


    def requestBanlist(self):
        """Request an updated banlist."""
        self._send_command('blocklist', 'block', '', 'next', '500')


    def requestUnBanlist(self):
        """Request an updated banlist."""
        self._send_command('blocklist', 'unblock', '', 'next', '500')


    def rawUnban(self, name, ip, unid):
        """
        Execute the unblock command using specified arguments.
        (For advanced usage)

        @type name: str
        @param name: name
        @type ip: str
        @param ip: ip address
        @type unid: str
        @param unid: unid
        """
        self._send_command('removeblock', unid, ip, name)


    def unban(self, user):
        """
        Unban a user. (Moderator only)

        @type user: User
        @param user: user to unban

        @rtype: bool
        @return: whether it succeeded
        """
        rec = self._getBanRecord(user)
        if rec:
            self.rawUnban(rec['target'].name, rec['ip'], rec['unid'])
            return True
        else:
            return False


    def _getBanRecord(self, user):
        return self._banlist.get(user)


    def getLevel(self, user):
        """get the level of user in a room"""
        if user == self.owner:
            return ROOM_OWNER
        if user.name in self.modnames:
            return ROOM_MODERATOR
        return 0


    def getLastMessage(self, user=None):
        """get last message said by user in a room"""
        if user:
            try:
                i = 1
                while True:
                    msg = self._history[-i]
                    if msg.user == user:
                        return msg
                    i += 1
            except IndexError:
                return None
        else:
            try:
                return self._history[-1]
            except IndexError:
                return None
        return None


    def findUser(self, name):
        """check if user is in the room

        return User(name) if name in room else None"""
        name = name.lower()
        udi = {user.name.lower(): user for user in self._userlist}
        cname = None
        for n in udi.keys():
            if name in n:
                if cname:
                    return None #ambiguous!!
                cname = n
        if cname:
            return udi[cname]
        else:
            return None


    def _addHistory(self, msg):
        """
        Add a message to history.

        @type msg: Message
        @param msg: message
        """
        self._history.append(msg)
        if len(self._history) > conf['history']['size']:
            rest, self._history = (self._history[:-conf['history']['size']],
                    self._history[-conf['history']['size']:])

            for msg in rest:
                msg.detach()


    @staticmethod
    def _getAnonId(n, ssid):
        """Gets the anon's id."""
        if n is None:
            n = '5504'
        try:
            return ''.join(str(x[0] + x[1])[-1] for x in zip(
                        [int(x) for x in n],
                        [int(x) for x in ssid[4:]])
                    )

        except ValueError:
            return 'NNNN'


    @staticmethod
    def _parseNameColor(n):
        """This just returns its argument, should return the name color."""
        #probably is already the name
        return n


    @staticmethod
    def _parseFont(f):
        """Parses the contents of a f tag and returns color, face and size."""
        #' xSZCOL="FONT"'
        try: #TODO: remove quick hack
            # sizeColor, fontFace
            sizecolor, _ = f.split('=', 1)
            sizecolor = sizecolor.strip()
            size = int(sizecolor[1:3])
            col = sizecolor[3:6]
            if col == '':
                col = None
            face = f.split('"', 2)[1]
            return col, face, size
        except: # pylint: disable=bare-except
            return None, None, None


    ## Received commands


    @asyncio.coroutine
    def _rcmd_ok(self, args):
        try:
            username, password = self.mgr.get_user_credentials()

            if password is None:
                # if got name, join room as name and no password
                if args[2] == "N":
                    self.login(username)

            elif args[2] != 'M':
                # unsuccessful login
                self._call_event('onLoginFail')
                self.disconnect()

        except ValueError:
            # if no name, join room as anon and no password
            if args[2] == 'N':
                n = args[4].rsplit('.', 1)[0]
                n = n[-4:]
                aid = args[1][0:8]
                pid = '!anon' + self._getAnonId(n, aid)
                self._botname = pid
                self.user.nameColor = n

        self.owner = self.user_class.create(args[0])
        self._uid = args[1]
        self._aid = args[1][4:8]
        self.mods.clear()
        for name in args[6].split(';'):
            self.mods.add(self.user_class.create(name.split(',')[0]))

        self._i_log.clear()


    @asyncio.coroutine
    def _rcmd_denied(self, args):
        self.disconnect()
        self._call_event('onConnectFail')


    @asyncio.coroutine
    def _rcmd_inited(self, args):
        self._send_command('g_participants', 'start')
        self._send_command('getpremium', '1')
        self.requestBanlist()
        self.requestUnBanlist()
        if self._connectAmmount == 0:
            self._call_event('onConnect')
            for msg in reversed(self._i_log):
                user = msg.user
                self._call_event('onHistoryMessage', user, msg)
                self._addHistory(msg)
            self._i_log.clear()
        else:
            self._call_event('onReconnect')
        self._connectAmmount += 1


    @asyncio.coroutine
    def _rcmd_premium(self, args):
        if float(args[1]) > time():
            self._premium = True
            if self.user._mbg:
                self.setBgMode(1)
            if self.user._mrec:
                self.setRecordingMode(1)
        else:
            self._premium = False


    @asyncio.coroutine
    def _rcmd_mods(self, args):
        modnames = args
        mods = set()
        for name in modnames:
            mods.add(self.user_class.create(name.split(',')[0]))

        premods = self.mods
        for user in mods - premods: #modded
            self.mods.add(user)
            self._call_event('onModAdd', user)

        for user in premods - mods: #demodded
            self.mods.remove(user)
            self._call_event('onModRemove', user)

        self._call_event('onModChange')


    @asyncio.coroutine
    def _rcmd_b(self, args):
        mtime = float(args[0])
        puid = args[3]
        ip = args[6]
        name = args[1]
        rawmsg = ':'.join(args[9:])
        msg, n, f = self.clean_message(rawmsg)
        if name == '':
            nameColor = None
            name = '#' + args[2]
            if name == '#':
                name = '!anon' + self._getAnonId(n, puid)
        else:
            if n:
                nameColor = self._parseNameColor(n)
            else:
                nameColor = None

        i = args[5]
        unid = args[4]
        user = self.user_class.create(name)
        if puid:
            user.puid = puid

        #Create an anonymous message and queue it because msgid is unknown.
        if f:
            fontColor, fontFace, fontSize = self._parseFont(f)
        else:
            fontColor, fontFace, fontSize = None, None, None

        msg = self.message_class(time=mtime, user=user, body=msg, raw=rawmsg,
                ip=ip, nameColor=nameColor, fontColor=fontColor,
                fontFace=fontFace, fontSize=fontSize, unid=unid, room=self)

        self._mqueue[i] = msg


    @asyncio.coroutine
    def _rcmd_u(self, args):
        temp = Struct(**self._mqueue)
        if hasattr(temp, args[0]):
            msg = getattr(temp, args[0])
            if msg.user != self.user:
                msg.user.fontColor = msg.fontColor
                msg.user.fontFace = msg.fontFace
                msg.user.fontSize = msg.fontSize
                msg.user.nameColor = msg.nameColor

            del self._mqueue[args[0]]
            msg.attach(self, args[1])
            self._addHistory(msg)
            self._call_event('onMessage', msg.user, msg)


    @asyncio.coroutine
    def _rcmd_i(self, args):
        mtime = float(args[0])
        puid = args[3]
        ip = args[6]
        name = args[1]
        rawmsg = ':'.join(args[9:])
        msg, n, f = self.clean_message(rawmsg)
        if name == '':
            nameColor = None
            name = '#' + args[2]
            if name == '#':
                name = '!anon' + self._getAnonId(n, puid)
        else:
            if n:
                nameColor = self._parseNameColor(n)
            else:
                nameColor = None

        #i = args[5]
        unid = args[4]
        user = self.user_class.create(name)
        if puid:
            user.puid = puid

        #Create an anonymous message and queue it because msgid is unknown.
        if f:
            fontColor, fontFace, fontSize = self._parseFont(f)
        else:
            fontColor, fontFace, fontSize = None, None, None

        msg = self.message_class(time=mtime, user=user, body=msg, raw=rawmsg,
                ip=ip, nameColor=nameColor, fontColor=fontColor,
                fontFace=fontFace, fontSize=fontSize, unid=unid, room=self)

        self._i_log.append(msg)


    @asyncio.coroutine
    def _rcmd_g_participants(self, args):
        args = ':'.join(args)
        args = args.split(';')
        for data in args:
            data = data.split(':')
            name = data[3].lower()
            if name == 'none':
                continue
            user = self.user_class.create(name=name, room=self)
            user.addSessionId(self, data[0])
            self._userlist.append(user)
        self._call_event('onUserList', self._userlist)


    @asyncio.coroutine
    def _rcmd_participant(self, args):
        name = args[3].lower()
        if name == "none":
            return
        user = self.user_class.create(name)
        puid = args[2]
        if puid:
            user.puid = puid

        if args[0] == '0': #leave
            user.removeSessionId(self, args[1])
            try:
                # last_on, is_on, idle
                _, _, _ = self._status[user]
            except: # pylint: disable=bare-except
                pass

            #if user not in self._userlist or not self.mgr._userlistEventUnique:
            self._call_event('onLeave', user)

        else: #join
            user.addSessionId(self, args[1])
            #doEvent = user not in self._userlist
            self._userlist.append(user)
            #if doEvent or not self.mgr._userlistEventUnique:
            self._call_event('onJoin', user)


    @asyncio.coroutine
    def _rcmd_show_fw(self, args):
        self._call_event('onFloodWarning')


    @asyncio.coroutine
    def _rcmd_show_tb(self, args):
        self._call_event('onFloodBan')


    @asyncio.coroutine
    def _rcmd_tb(self, args):
        self._call_event('onFloodBanRepeat')


    @asyncio.coroutine
    def _rcmd_delete(self, args):
        msg = self._msgs.get(args[0])
        if msg:
            if msg in self._history:
                self._history.remove(msg)
                self._call_event('onMessageDelete', msg.user, msg)
                msg.detach()


    @asyncio.coroutine
    def _rcmd_deleteall(self, args):
        for msgid in args:
            yield from self._rcmd_delete([msgid])


    @asyncio.coroutine
    def _rcmd_n(self, args):
        self.usercount = int(args[0], 16)
        self._call_event('onUserCountChange')


    @asyncio.coroutine
    def _rcmd_blocklist(self, args):
        self._banlist = dict()
        sections = ':'.join(args).split(';')
        for section in sections:
            params = section.split(":")
            if len(params) != 5:
                continue
            if params[2] == '':
                continue

            user = self.user_class.create(params[2])
            self._banlist[user] = {
                'unid': params[0],
                'ip': params[1],
                'target': user,
                'time': float(params[3]),
                'src': self.user_class.create(params[4]),
            }
        self._call_event('onBanlistUpdate')


    @asyncio.coroutine
    def _rcmd_unblocklist(self, args):
        self._unbanlist = dict()
        sections = ':'.join(args).split(';')
        for section in sections:
            params = section.split(':')
            if len(params) != 5:
                continue
            if params[2] == '':
                continue

            user = self.user_class.create(params[2])
            self._unbanlist[user] = {
                'unid': params[0],
                'ip': params[1],
                'target': user,
                'time': float(params[3]),
                'src': self.user_class.create(params[4]),
            }
        self._call_event('onUnBanlistUpdate')


    @asyncio.coroutine
    def _rcmd_blocked(self, args):
        if args[2] == '':
            return
        target = self.user_class.create(args[2])
        user = self.user_class.create(args[3])
        self._banlist[target] = {
            'unid': args[0],
            'ip': args[1],
            'target': target,
            'time': float(args[4]),
            'src': user,
        }
        self._call_event('onBan', user, target)


    @asyncio.coroutine
    def _rcmd_unblocked(self, args):
        if args[2] == r'':
            return
        target = self.user_class.create(args[2])
        user = self.user_class.create(args[3])
        del self._banlist[target]
        self._unbanlist[user] = {
            'unid': args[0],
            'ip': args[1],
            'target': target,
            'time': float(args[4]),
            'src': user,
        }
        self._call_event('onUnban', user, target)



import asyncio
import os
import re
from datetime import datetime
from email.utils import parsedate_tz, mktime_tz
from pytz import utc


def coro_later(loop, delay, coroutine):
    loop.call_later(delay, lambda: asyncio.ensure_future(coroutine))


def load_module(basedir, package, module_name):
    '''Load a python module by its name.

    See: http://stackoverflow.com/a/67692
    '''
    base_dir = os.environ['PROJECT_DIR']
    module_path = os.path.join(base_dir, package, module_name + '.py')
    module_name = package + '.' + module_name
    try:
        # python 3.5
        from importlib.util import spec_from_file_location, module_from_spec
        spec = spec_from_file_location(module_name, module_path)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
    except ImportError:
        # python 3.4
        try:
            from importlib.machinery import SourceFileLoader
            module = SourceFileLoader(module_name, module_path).load_module()
        except ImportError:
            # python 2
            from imp import load_source
            module = load_source(module_name, module_path)

    return module


def create_word_regex(word):
    return re.compile(r'(^|[\s?!.,;])' + word + r'($|[\s?!.,;])')


def parse_date(text):
    timestamp = mktime_tz(parsedate_tz(text))
    return datetime.fromtimestamp(timestamp, utc)


def format_date(date):
    return date.strftime('%a, %d %b %Y %H:%M:%S GMT')



from setuptools import setup, find_packages

RELEASE_VERSION = '0.0.6'

setup(
    name='pychatangobot',
    version=RELEASE_VERSION,
    url='https://github.com/dozymoe/PyChatangoBot',
    download_url='https://github.com/dozymoe/PyChatangoBot/tarball/' +\
            RELEASE_VERSION,

    author='Fahri Reza',
    author_email='dozymoe@gmail.com',
    description='asyncio implementation of Chatango bot (ch.py).',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    platforms='any',
    license='MIT',
    install_requires=[
        "aiohttp",
        "bleach",
        "pycircularbuffer",
        "pytz",
    ],
    ext_modules=[],
)

print('Visit https://github.com/dozymoe/PyChatangoBot for example ' +\
        'run-bot scripts.')



import asyncio

from re import search
from uuid import uuid4
from html import escape, unescape
from logging import getLogger
from circularbuffer import CircularBuffer # pylint:disable=no-name-in-module
from bleach import clean

from .settings import conf
from ..utils import coro_later

class BaseChannel(asyncio.Protocol):
    """Manages chatroom."""

    connected = False
    mgr = None
    name = None

    event_name_ping = 'onPing'

    _buf = None
    _cid = None # Connection Id
    _conn = None
    _log = None
    _loop = None
    _future = None

    _firstCommand = True


    def clean_message(self, html):
        n = search(r'<n(.*?)/>', html)
        if n:
            n = n.group(1)

        f = search(r'<f(.*?)>', html)
        if f:
            f = f.group(1)

        return (unescape(clean(html, strip=True)), n, f)


    def html_escape(self, text, quote=False):
        return escape(text, quote)


    def strip_tags(self, text):
        return clean(text, strip=True)


    def get_server(self):
        raise NotImplementedError()


    @asyncio.coroutine
    def authenticate(self):
        raise NotImplementedError()


    def _disconnect(self):
        pass


    @asyncio.coroutine
    def connect(self):
        """Connect or reconnect to server.
        see: http://stackoverflow.com/a/26004654
        """
        host, port = self.get_server()
        retries = conf['connection']['max_retries']
        retry_delay = conf['connection']['retry_delay']
        self.connected = True
        while retries > 0 and self.connected:
            retries -= 1
            try:
                self._cid = uuid4().int
                self._buf.clear()

                yield from self._loop.create_connection(self, host, port)
                if self._future is None:
                    try:
                        self._future = self._loop.create_future()
                    except AttributeError:
                        self._future = asyncio.Future()
                return

            except OSError:
                self._log.info('Failed to connect, retrying in %i seconds...',
                        retry_delay)

                yield from asyncio.sleep(retry_delay)
            except: # pylint: disable=bare-except
                break

        self.connected = False


    def disconnect(self):
        """Disconnect from server.

        Keep this function idempotent, can be called multiple times without
        having to check if certain flags were set.
        """
        future = self._future

        # This will cancel reconnects
        self.connected = False

        # This will cancel self._ping infinite-loop
        self._cid = None

        if self._conn is not None:
            self._conn.close()

        # Connection was never made, create false Future
        if future is None:
            future = asyncio.Future()
            future.set_result(None)

        return future


    def connection_made(self, transport):
        self._conn = transport
        self._firstCommand = True

        asyncio.ensure_future(self.authenticate())
        asyncio.ensure_future(self._ping(self._cid))


    def connection_lost(self, exc):
        self._conn = None
        if exc is not None:
            self._log.warning(repr(exc))

        if not self.connected:
            self._disconnect()
            self._future.set_result(None)
            self._future = None
            return

        coro_later(self._loop, conf['connection']['retry_delay'],
                self.connect())


    def data_received(self, data):
        total = len(data)
        wrote = 0
        while wrote < total:
            count = self._buf.write(data[wrote:])
            wrote += count

            while True:
                try:
                    pos = self._buf.index(b'\0')
                except ValueError:
                    break

                recv = self._buf.read(pos)
                # drop the \0
                self._buf.read(1)

                recv = recv.decode('utf-8', errors='replace').rstrip('\r\n')
                asyncio.ensure_future(self._process(recv))


    @asyncio.coroutine
    def _process(self, recv):
        """Process a command string.

        @type data: str
        @param data: the command string
        """
        self._call_event('onRaw', recv)

        data = recv.split(':')
        cmd, args = data[0], data[1:]

        func = '_rcmd_' + cmd
        if hasattr(self, func):
            yield from getattr(self, func)(args)
        elif len(recv):
            self._log.warning('Unknown data received: ' + repr(recv))


    def _call_event(self, name, *args, **kwargs):
        asyncio.ensure_future(getattr(self.mgr, name)(self, *args, **kwargs))
        asyncio.ensure_future(self.mgr.onEventCalled(self, name, *args,
                **kwargs))


    def _send_command(self, *args):
        """
        Send a command.

        @type args: [str, str, ...]
        @param args: command and list of arguments
        """
        if self._conn is None:
            return

        if self._firstCommand:
            terminator = b'\x00'
            self._firstCommand = False
        else:
            terminator = b'\r\n\x00'

        self._conn.write(':'.join(args).encode('utf-8') + terminator)


    @asyncio.coroutine
    def _ping(self, unique_id):
        """Send a ping."""
        interval = conf['connection']['ping_interval']
        count = interval / 10
        if interval % 10:
            count += 1

        while count > 0:
            count -= 1
            yield from asyncio.sleep(10)
            if self._cid != unique_id:
                return

        self._send_command('')
        self._call_event(self.event_name_ping)

        asyncio.ensure_future(self._ping(unique_id))


    def __call__(self):
        """Dummy function to be use as ProtocolFactory for asyncio."""
        return self


    def __init__(self, loop, mgr):
        self.mgr = mgr
        self._loop = loop
        self._buf = CircularBuffer(conf['connection']['buffer_size'])

        log_name = type(self).__name__
        if self.name is not None:
            log_name += '(%s)' % self.name

        self._log = getLogger(log_name)



"""Chat room users."""

_users = {}

class User(object):
    """Class that represents a user."""

    name = None
    puid = ''
    fontColor = '000'
    fontFace = '0'
    fontSize = 12
    nameColor = '000'

    _sids = None
    _msgs = None
    _mbg = False
    _mrec = False

    @property
    def sessionIds(self):
        return self._getSessionIds()


    @property
    def rooms(self):
        return self._sids.keys()


    @property
    def roomnames(self):
        return [room.name for room in self.rooms]


    @classmethod
    def create(cls, name, *args, **kwargs):
        if name is None:
            name = ''

        name = name.lower()
        user = _users.get(name)
        if not user:
            user = cls(name=name, *args, **kwargs)
            _users[name] = user
        return user


    def addSessionId(self, room, sid):
        if room not in self._sids:
            self._sids[room] = set()
        self._sids[room].add(sid)


    def removeSessionId(self, room, sid):
        try:
            self._sids[room].remove(sid)
            if len(self._sids[room]) == 0:
                del self._sids[room]
        except KeyError:
            pass


    def clearSessionIds(self, room):
        try:
            del self._sids[room]
        except KeyError:
            pass


    def hasSessionId(self, room, sid):
        try:
            return sid in self._sids[room]
        except KeyError:
            return False


    def _getSessionIds(self, room=None):
        if room:
            return self._sids.get(room, set())
        else:
            return set.union(*self._sids.values())


    def __repr__(self):
        return "<User: %s>" % self.name


    def __init__(self, name, **kw):
        self.name = name.lower()
        self._sids = {}
        self._msgs = []

        for attr, val in kw.items():
            if val is not None:
                setattr(self, attr, val)
